# -*- coding: utf-8 -*-


import cjieba

from ..solo import singleton
from ..summarization import STOPWORDS


def load_stopwords(stopwords_path):
    """
    加载停用词词典
    :param stopwords_path:
    :return:
    """
    with open(stopwords_path, encoding='utf-8') as f:
        return [line.strip() for line in f]


def split_doc(doc):
    separators = ['。', '!', '！', '？', '?']
    for sep in separators:
        doc = doc.replace(sep, sep + '##')
    sentences = doc.split('##')
    return sentences[:-1]


def calculate_sentence_score(sentence, stopwords):
    # jieba_ret = jieba.analyse.extract_tags(sentence, topK=100, withWeight=True)  # , allowPOS=('ns', 'n', 'vn', 'v'))
    jieba_ret = cjieba.extract(sentence, top_k=100, with_weight=True)
    sentence_score = 0
    for word, score in jieba_ret:
        if word not in stopwords:
            sentence_score += score
    return sentence_score


@singleton
class TfidfSummarization:
    def __init__(self, ratio=0.2):
        self.ratio = ratio

    def analysis(self, doc):
        stopwords = load_stopwords(STOPWORDS)

        # 对文本进行分割
        sentences = split_doc(doc)

        # 存放句子顺序
        sentences_order = {sent: idx for idx, sent in enumerate(sentences)}

        # 存放句子得分
        sentences_score = {sent: calculate_sentence_score(sent, stopwords) for sent in sentences}

        # 根据得分，选出较高得分的句子
        sentences_score_order = sorted(sentences_score.items(), key=lambda item: -item[1])[
                                : int(len(sentences) * self.ratio)]

        # 将较高的分的句子按原文本进行排序输出
        selected_sentences = {sent: sentences_order[sent] for sent, score in sentences_score_order}
        summary = ''.join([sent for sent, order in sorted(selected_sentences.items(), key=lambda item: item[1])])

        return summary


if __name__ == '__main__':
    doc = """在主持学习时发表了讲话。他强调，人工智能是引领这一轮科技革命和产业变革的战略性技术，具有溢出带动性很强的“头雁”效应。在移动互联网、大数据、超级计算、传感网、脑科学等新理论新技术的驱动下，人工智能加速发展，呈现出深度学习、跨界融合、人机协同、群智开放、自主操控等新特征，正在对经济发展、社会进步、国际政治经济格局等方面产生重大而深远的影响。加快发展新一代人工智能是我们赢得全球科技竞争主动权的重要战略抓手，是推动我国科技跨越发展、产业优化升级、生产力整体跃升的重要战略资源。习近平指出，人工智能具有多学科综合、高度复杂的特征。我们必须加强研判，统筹谋划，协同创新，稳步推进，把增强原创能力作为重点，以关键核心技术为主攻方向，夯实新一代人工智能发展的基础。要加强基础理论研究，支持科学家勇闯人工智能科技前沿的“无人区”，努力在人工智能发展方向和理论、方法、工具、系统等方面取得变革性、颠覆性突破，确保我国在人工智能这个重要领域的理论研究走在前面、关键核心技术占领制高点。要主攻关键核心技术，以问题为导向，全面增强人工智能科技创新能力，加快建立新一代人工智能关键共性技术体系，在短板上抓紧布局，确保人工智能关键核心技术牢牢掌握在自己手里。要强化科技应用开发，紧紧围绕经济社会发展需求，充分发挥我国海量数据和巨大市场应用规模优势，坚持需求导向、市场倒逼的科技发展路径，积极培育人工智能创新产品和服务，推进人工智能技术产业化，形成科技创新和产业应用互相促进的良好发展局面。要加强人才队伍建设，以更大的决心、更有力的措施，打造多种形式的高层次人才培养平台，加强后备人才培养力度，为科技和产业发展提供更加充分的人才支撑。"""



# -*- coding: utf-8 -*-
import re


class LangDetector(object):
    """
    检测语言，传入字符串初始化，也可以不传字符串
    """
    # 不能检测出语言的字符
    # 先清理这些字符，再用正则判断unicode判断语言
    remove_characters = u'''[’·°–!"#$%&\'()*+,-./:;<=>?@，。?★、…【】（）《》？“”‘’！[\\]^_`{|}~0-9]+'''
    lang_pattern_list = {
        'zh': re.compile(u"[\u4e00-\u9fff]"),
        'en': re.compile(u'[a-z][A-Z]'),
        'ja': re.compile(u'[\u30a0-\u30ff\u3040-\u309f]'),
        'ko': re.compile(u'[\uac00-\ud7af\u1100-\u11ff\u3130-\u318f]')
    }
    # 默认使用的语言，识别不了的字符串会默认当作默认语言的字符
    default_lang = 'zh'

    # def __init__(self, text):
    # self.text = text

    def detect(self, text=None, cleaning=True, specific=False, cleaningExclude=False, unknownUseDefault=True):
        # text是不可变对象，因此可以传参
        # if text is None:
        # text = self.text
        if cleaning:
            if cleaningExclude:
                text = self.cleaning_exclude(text)
            else:
                text = self.cleaning_text(text)
        resDict = {}
        # 初始化语言计数用字典
        for lang in self.lang_pattern_list.keys():
            resDict[lang] = 0
        unknown_list = []
        for c in text:
            # 判断语言匹配列表是否匹配成功，如果没有匹配成功，
            # 加入到未知字符列表
            match_flag = False
            for lang, pattern in self.lang_pattern_list.items():
                if pattern.match(c):
                    match_flag = True
                    resDict[lang] += 1
            # 如果匹配不成功，加入未知字符列表
            if not match_flag:
                unknown_list.append(c)

        # 判断是否把未知字符作为默认语言的字符
        if unknownUseDefault:
            for _ in unknown_list:
                resDict[self.default_lang] += 1
            # 清空未知字符列表
            unknown_list.clear()
        result_list = []
        for lang, count in resDict.items():
            if count > 0:
                result_list.append([lang, count / len(text)])
        unknown_num = len(unknown_list)
        if unknown_num > 0:
            result_list.append(['unknown', unknown_num / len(text)])
        # 进行排序
        self.sort_lang_list(result_list)
        # 判断是否返回一个确定的值，默认返回的是各种语言字符串占比的列表
        if specific:
            return result_list[0][0]
        return result_list

    def cleaning_text(self, text):
        return re.sub(self.remove_characters, '', text)

    def cleaning_exclude(self, text):
        patternStr = self.exclude_lang_pattern()
        return re.sub(patternStr, '', text)

    def sort_lang_list(self, lang_list):
        """
        排序的字符串的格式是[['zh', 0.35714285714285715], ['ja', 0.6428571428571429]]
        """

        def sort_key(x):
            return x[1]

        lang_list.sort(key=sort_key, reverse=True)

    def exclude_lang_pattern(self):
        concatStr = ''
        for pattern in self.lang_pattern_list.values():
            concatStr += pattern.pattern
        tempstr = concatStr.replace('[', '').replace(']', '')
        return f'[^{tempstr}]'



# -*- coding: utf-8 -*-

import cjieba

from ..solo import singleton


@singleton
class TfidfKeywords:
    def __init__(self, delete_stopwords=True, topK=20, withWeight=False):
        self.topk = topK
        self.with_wight = withWeight

    def keywords(self, sentence):
        return cjieba.extract(text=sentence, top_k=self.topk, with_weight=self.with_wight)
        # return jieba.analyse.extract_tags(sentence, topK=self.topk, withWeight=self.with_wight)



# -*- coding: utf-8 -*-
# @Time    : 2023/9/6 下午4:58
# @Author  : sudoskys
# @File    : solo.py
# @Software: PyCharm
def singleton(cls):
    _instance = {}

    def _singleton(*args, **kargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kargs)
        return _instance[cls]

    return _singleton



# -*- coding: utf-8 -*-

from pathlib import Path
from types import ModuleType
from typing import Optional, Set, Iterable, Union

from . import (
    _managers,
    _current_plugin_chain,
    _module_name_to_plugin_name,
    path_to_module_name, _find_manager_by_name, get_plugin,
)
from .model import PluginManager
from .schema import Plugin


def load_all_plugins(module_path: Iterable[str], plugin_dir: Iterable[str]) -> Set[Plugin]:
    """
    导入指定列表中的插件以及指定目录下多个插件，以 `_` 开头的插件不会被导入!
    :param module_path: 指定插件集合
    :param plugin_dir: 指定文件夹路径集合
    """
    manager = PluginManager(module_path, plugin_dir)
    _managers.append(manager)
    return manager.load_all_plugins()


def load_plugins(*plugin_dir: str) -> Set[Plugin]:
    """
    导入文件夹下多个插件，以 `_` 开头的插件不会被导入!
    :param plugin_dir: 文件夹路径
    :return: 插件集合
    """
    manager = PluginManager(search_path=plugin_dir)
    _managers.append(manager)
    return manager.load_all_plugins()


def load_builtin_plugin(name: str) -> Optional[Plugin]:
    """导入 Bot 内置插件。
    :param name: 插件名称
    :return: 插件
    """
    return load_plugin(f"extra.plugins.{name}")


def load_builtin_plugins(*plugins: str) -> Set[Plugin]:
    """导入多个 Bot 内置插件。
    :param plugins: 插件名称集合
    """
    return load_all_plugins([f"extra.plugins.{p}" for p in plugins], [])


def load_plugin(module_path: Union[str, Path]) -> Optional[Plugin]:
    """
    加载单个插件，可以是本地插件或是通过 `pip` 安装的插件。
    :param module_path: 插件名称 `path.to.your.plugin`
            或插件路径 `pathlib.Path(path/to/your/plugin)`
    :return: 插件
    """
    module_path = (
        path_to_module_name(module_path)
        if isinstance(module_path, Path)
        else module_path
    )
    manager = PluginManager([module_path])
    _managers.append(manager)
    return manager.load_plugin(module_path)


def load_from_entrypoint(group="llmkira.extra.plugin") -> Set[Plugin]:
    import importlib_metadata
    hook = importlib_metadata.entry_points().select(group=group)
    plugins = [item.module for item in hook]
    return load_all_plugins(plugins, [])


def require(name: str) -> ModuleType:
    """
    获取一个插件的导出内容。
    如果为 `load_plugins` 文件夹导入的插件，则为文件(夹)名。
    :param name: 插件名称 即 {ref}`extra.plugin.model.Plugin.name`。
    :exception RuntimeError: 插件无法加载
    :return: 插件导出内容
    """
    plugin = get_plugin(_module_name_to_plugin_name(name))
    # if plugin not loaded
    if not plugin:
        # plugin already declared
        manager = _find_manager_by_name(name)
        if manager:
            plugin = manager.load_plugin(name)
        # plugin not declared, try to declare and load it
        else:
            # clear current plugin chain, ensure plugin loaded in a new context
            _t = _current_plugin_chain.set(())
            try:
                plugin = load_plugin(name)
            finally:
                _current_plugin_chain.reset(_t)
    if not plugin:
        raise RuntimeError(f'Cannot load plugin "{name}"!')
    return plugin.module



# -*- coding: utf-8 -*-
# @Time    : 2023/8/21 上午12:06
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from llmkira.cache.redis import cache
from llmkira.utils import sync
from loguru import logger

from .schema import UserInfo
from ...sdk.endpoint import openai


class SubManager(object):
    def __init__(self, user_id: str):
        self.reload_exception = None
        self.sub_info: UserInfo = sync(self._sync(user_id=user_id))
        assert isinstance(self.sub_info, UserInfo), "sub info error"

    async def _upload(self, user_id: str, ignore_exception=False):
        if self.reload_exception and not ignore_exception:
            logger.warning(f"you are upload a default sub info for {user_id} to wipe out the original data")
            return None
        await cache.set_data(key=f"sub:{user_id}", value=self.sub_info.json())

    async def _sync(self, user_id: str):
        _cache = await cache.read_data(key=f"sub:{user_id}")
        if not _cache:
            return UserInfo(user_id=user_id)
        try:
            sub_info = UserInfo().parse_obj(_cache)
        except Exception as e:
            self.reload_exception = True
            logger.warning(f"not found sub info {user_id},{e}")
            return UserInfo(user_id=user_id)
        return sub_info

    @property
    def llm_driver(self):
        if not self.sub_info.update_driver:
            return openai.Openai.Driver()
        return self.sub_info.llm_driver

    async def block_plugin(self, plugin_name: str):
        self.sub_info.plugin_subs.lock.append(plugin_name)
        return self.sub_info.plugin_subs.lock

    async def unblock_plugin(self, plugin_name: str):
        if plugin_name in self.sub_info.plugin_subs.lock:
            self.sub_info.plugin_subs.lock.remove(plugin_name)
        return self.sub_info.plugin_subs.lock

    async def set_endpoint(self, endpoint: str = None, api_key: str = None):
        """
        这里容易产生鉴权错误和环境变量被泄漏的问题！！
        # TODO: 优化鉴权机制
        """
        if not endpoint:
            endpoint = "https://api.openai.com/v1/chat/completions"
        self.sub_info.llm_driver = openai.Openai.Driver(endpoint=endpoint, api_key=api_key)
        self.sub_info.update_driver = True
        await self._upload(user_id=self.sub_info.user_id)

    async def clear_endpoint(self):
        self.sub_info.update_driver = False
        self.sub_info.llm_driver = None
        await self._upload(user_id=self.sub_info.user_id)

    async def get_lock_plugin(self):
        return self.sub_info.plugin_subs.lock

    async def add_cost(self, cost: UserInfo.Cost) -> list:
        self.sub_info.costs.append(cost)
        await self._upload(user_id=self.sub_info.user_id)
        return self.sub_info.costs

    async def get_total_cost(self) -> int:
        return self.sub_info.total_cost()

    async def get_cost_by_function_name(self, function_name: str) -> list:
        return [cost for cost in self.sub_info.costs if cost.function_name == function_name]



# -*- coding: utf-8 -*-
# @Time    : 2023/10/14 下午2:17

from llmkira.sdk.func_calling import get_available_plugin_names as get_available_plugin_names
from llmkira.sdk.func_calling import get_loaded_plugins as get_loaded_plugins
from llmkira.sdk.func_calling import get_plugin as get_plugin
from llmkira.sdk.func_calling import load_all_plugins as load_all_plugins
from llmkira.sdk.func_calling import load_builtin_plugin as load_builtin_plugin
from llmkira.sdk.func_calling import load_builtin_plugins as load_builtin_plugins
from llmkira.sdk.func_calling import load_plugin as load_plugin
from llmkira.sdk.func_calling import load_plugins as load_plugins
from llmkira.sdk.func_calling import require as require



# -*- coding: utf-8 -*-
# @Time    : 2023/10/11 下午3:35
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from typing import Any, Type

import shortuuid
from pydantic import BaseModel, Field, validator


class Chain(BaseModel):
    user_id: str = Field(None, description="user_id")
    address: str = Field(None, description="address")
    time: int = 0
    arg: Any = Field(None, description="arg")
    uuid: str = shortuuid.uuid()[0:8]

    @validator("user_id")
    def check_user_id(cls, v):
        if not v:
            raise ValueError("Chain:user_id is empty")
        return v

    @validator("address")
    def check_address(cls, v):
        if not v:
            raise ValueError("Chain:address is empty")
        return v

    def format_arg(self, arg: Type[BaseModel]):
        if isinstance(self.arg, dict):
            self.arg = arg.parse_obj(self.arg)
        return self



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:22
# @Author  : sudoskys
# @File    : telegram.py
# @Software: PyCharm
from dotenv import load_dotenv
from loguru import logger
from pydantic import BaseSettings, Field, validator, root_validator


class TelegramBot(BaseSettings):
    """
    代理设置
    """
    token: str = Field(None, env='TELEGRAM_BOT_TOKEN')
    proxy_address: str = Field(None, env="TELEGRAM_BOT_PROXY_ADDRESS")  # "all://127.0.0.1:7890"
    bot_link: str = Field(None, env='TELEGRAM_BOT_LINK')
    bot_id: str = Field(None, env="TELEGRAM_BOT_ID")
    bot_username: str = Field(None, env="TELEGRAM_BOT_USERNAME")

    class Config:
        env_file = '.env'
        env_file_encoding = 'utf-8'

    @validator('token')
    def proxy_address_validator(cls, v):
        if v is None:
            raise ValueError('Bot token is None')
        return v

    @root_validator
    def bot_id_validator(cls, values):
        if values.get('bot_id') is None:
            try:
                from telebot import TeleBot
                # 创建 Bot
                if values['proxy_address'] is not None:
                    from telebot import apihelper
                    if "socks5://" in values['proxy_address']:
                        values['proxy_address'] = values['proxy_address'].replace("socks5://", "socks5h://")
                    apihelper.proxy = {'https': values['proxy_address']}
                _bot = TeleBot(token=values.get('token')).get_me()
                values['bot_id'] = _bot.id
                values['bot_username'] = _bot.username
                values['bot_link'] = f"https://t.me/{values['bot_username']}"
            except Exception as e:
                logger.error(f"TelegramBot id is None, {e}")
            else:
                logger.success(f"TelegramBot connect success: {values.get('bot_username')}")
        return values

    @property
    def available(self):
        return self.token is not None


load_dotenv()
BotSetting = TelegramBot()



# -*- coding: utf-8 -*-
# @Time    : 2023/10/14 下午2:16
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
import copy
import time
from typing import Literal, Tuple, List, Union, Optional

from dotenv import load_dotenv
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint import openai
from llmkira.sdk.schema import File
from llmkira.utils import sync
from loguru import logger
from pydantic import BaseSettings, Field, BaseModel, root_validator
from telebot import types


class RabbitMQ(BaseSettings):
    """
    代理设置
    """
    amqp_dsn: str = Field("amqp://admin:admin@localhost:5672", env='AMQP_DSN')
    _verify_status: bool = Field(False, env='VERIFY_STATUS')

    class Config:
        env_file = '.env'
        env_file_encoding = 'utf-8'

    @root_validator()
    def is_connect(cls, values):
        import aio_pika
        try:
            sync(aio_pika.connect_robust(
                values['amqp_dsn']
            ))
        except Exception as e:
            values['_verify_status'] = False
            logger.warning(f'RabbitMQ DISCONNECT, pls set AMQP_DSN in .env\n--error {e}')
        else:
            values['_verify_status'] = True
            logger.success(f"RabbitMQ connect success")
        return values

    def check_connection(self, values):
        import aio_pika
        try:
            sync(aio_pika.connect_robust(
                self.amqp_dsn
            ))
        except Exception as e:
            logger.warning('RabbitMQ DISCONNECT, pls set AMQP_DSN in ENV')
            raise ValueError('RabbitMQ connect failed')
        else:
            logger.success(f"RabbitMQ connect success")
        return values

    @property
    def task_server(self):

        return self.amqp_dsn


load_dotenv()
RabbitMQSetting = RabbitMQ()


class TaskHeader(BaseModel):
    """
    任务链节点
    """

    class Meta(BaseModel):
        class Callback(BaseModel):
            role: Literal["user", "system", "function", "assistant"] = Field("user", description="角色")
            name: str = Field(None, description="功能名称", regex=r"^[a-zA-Z0-9_]+$")

        sign_as: Tuple[int, str, str] = Field((0, "root", "default"), description="签名")

        function_enable: bool = Field(False, description="功能开关")
        function_list: List[openai.Function] = Field([], description="功能列表")
        function_salvation_list: List[openai.Function] = Field([], description="上回合的功能列表，用于容错")

        direct_reply: bool = Field(False, description="直接回复,跳过函数处理等")
        callback_forward: bool = Field(False, description="非 LLM 转发")
        callback_forward_reprocess: bool = Field(False, description="追加LLM回复,追加存储处理后再回复")
        continue_step: int = Field(0, description="继续执行的步骤,发送启用 function call 的 continue")
        limit_child: int = Field(4, description="限制子节点数量")

        verify_uuid: str = Field(None, description="用于验证重发")
        parent_call: openai.OpenaiResult = Field(None, description="存储上一个节点的父消息，用于插件的原始消息信息存储")
        callback: Callback = Field(Callback(), description="用于回写，插件返回的消息头，标识 function 的名字")
        extra_args: dict = Field({}, description="用于提供额外参数")

        class Config:
            arbitrary_types_allowed = True

        def child(self, name):
            self.sign_as = (self.sign_as[0] + 1, "child", name)
            self.limit_child -= 1
            return copy.deepcopy(self)

    class Location(BaseModel):
        platform: str = Field(None, description="平台")
        chat_id: int = Field(None, description="群组ID")
        user_id: int = Field(None, description="用户ID")
        message_id: int = Field(None, description="消息ID")

    class Plugin(BaseModel):
        name: str = Field(None, description="插件名称")
        is_run_out: bool = Field(False, description="是否运行完毕")
        token_usage: int = Field(0, description="Token 用量")

    task_meta: Meta = Field(Meta(), description="任务元数据")
    sender: Location = Field(None, description="发信人")
    receiver: Location = Field(None, description="接收人")
    message: List[RawMessage] = Field(None, description="消息内容")

    @classmethod
    def from_telegram(cls, message: Union[types.Message],
                      task_meta: Meta,
                      file: List[File] = None,
                      reply: bool = True,
                      hide_file_info: bool = False,
                      trace_back_message: List[types.Message] = None
                      ):
        """
        从telegram消息中构建任务
        """

        if (trace_back_message is None) or (not all(trace_back_message)):
            trace_back_message = []
        if file is None:
            file = []

        def _convert(_message: types.Message) -> Optional[RawMessage]:
            if not _message:
                raise ValueError(f"Message is empty")
            if isinstance(_message, types.Message):
                user_id = _message.from_user.id
                chat_id = _message.chat.id
                text = _message.text if _message.text else _message.caption
                created_at = _message.date
            else:
                raise ValueError(f"Unknown message type {type(_message)}")
            return RawMessage(
                user_id=user_id,
                chat_id=chat_id,
                text=text if text else f"(hi)",  # Empty Message
                created_at=created_at
            )

        _file_name = []
        if file:
            for _file in file:
                _file_name.append(_file.file_prompt)
        head_message = _convert(message)
        assert head_message, "HeadMessage is empty"
        head_message.file = file
        if not hide_file_info:
            head_message.text += "\n" + "\n".join(_file_name)
        message_list = []
        if trace_back_message:
            for item in trace_back_message:
                message_list.append(_convert(item))
        message_list.append(head_message)
        message_list = [item for item in message_list if item]
        # 去掉 None
        return cls(
            task_meta=task_meta,
            sender=cls.Location(
                platform="telegram",
                chat_id=message.chat.id,
                user_id=message.from_user.id,
                message_id=message.message_id if reply else None
            ),
            receiver=cls.Location(
                platform="telegram",
                chat_id=message.chat.id,
                user_id=message.from_user.id,
                message_id=message.message_id if reply else None
            ),
            message=message_list
        )

    @classmethod
    def from_function(cls,
                      parent_call: openai.OpenaiResult,
                      task_meta: Meta,
                      receiver: Location,
                      message: List[RawMessage] = None
                      ):
        """
        从 Openai LLM Task中构建任务
        'function_call': {'name': 'set_alarm_reminder', 'arguments': '{\n  "delay": "5",\n  "content": "该吃饭了"\n}'}}
        """
        task_meta.parent_call = parent_call
        return cls(
            task_meta=task_meta,
            sender=receiver,
            receiver=receiver,
            message=message
        )

    @classmethod
    def from_router(cls, from_, to_, user_id, method, message_text):
        _meta_arg = {}
        if method == "task":
            _meta_arg["function_enable"] = True
        elif method == "push":
            _meta_arg["callback_forward"] = True
        elif method == "chat":
            _meta_arg["function_enable"] = False
        meta = cls.Meta(
            **_meta_arg
        )
        return cls(
            task_meta=meta,
            sender=cls.Location(
                platform=from_,
                chat_id=user_id,
                user_id=user_id,
                message_id=None
            ),
            receiver=cls.Location(
                platform=to_,
                chat_id=user_id,
                user_id=user_id,
                message_id=None
            ),
            message=[RawMessage(
                user_id=user_id,
                chat_id=user_id,
                text=message_text,
                created_at=int(time.time())
            )]
        )



# -*- coding: utf-8 -*-
# @Time    : 2023/8/24 下午11:22
# @Author  : sudoskys
# @File    : search.py
# @Software: PyCharm
import os

from llmkira.middleware.chain_box import Chain, CHAIN_MANAGER
from llmkira.middleware.user import SubManager, UserInfo
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint import openai
from llmkira.sdk.endpoint.openai import Function
from llmkira.sdk.func_calling import BaseTool, PluginMetadata
from llmkira.sdk.func_calling.schema import FuncPair
from llmkira.sdk.schema import Message
from llmkira.task import Task, TaskHeader
from loguru import logger
from pydantic import BaseModel

__plugin_name__ = "search_in_google"

search = Function(name=__plugin_name__, description="Search/validate uncertain/unknownEvents/Meme fact on google.com")
search.add_property(
    property_name="keywords",
    property_description="question entered in the search website",
    property_type="string",
    required=True
)


def search_on_duckduckgo(search_sentence: str, key_words: str = None):
    logger.debug(f"Plugin:search_on_duckduckgo {search_sentence}")
    from duckduckgo_search import DDGS
    from llmkira.sdk.filter import Sublimate
    with DDGS(timeout=20) as ddgs:
        _text = []
        for r in ddgs.text(search_sentence):
            _title = r.get("title")
            _href = r.get("href")
            _body = r.get("body")
            _text.append(_body)
    if key_words:
        must_key = [key_words]
    else:
        must_key = None
    _test_result = Sublimate(_text).valuation(match_sentence=search_sentence, match_keywords=must_key)
    _result = []
    for key in _test_result[:4]:
        _result.append(key[0])
        # hidden clues
    return "\nHint Clues:".join(_result)


class Search(BaseModel):
    keywords: str

    class Config:
        extra = "allow"


class SearchTool(BaseTool):
    """
    搜索工具
    """
    silent: bool = False
    function: Function = search
    keywords: list = [
        "怎么", "How", "件事", "牢大", "作用", "知道", "什么", "认识", "What", "http",
        "what", "who", "how", "Who",
        "Why", "作品", "why", "Where",
        "了解", "简述", "How to", "是谁", "how to",
        "解释", "怎样的", "新闻", "ニュース", "电影", "番剧", "アニメ",
        "2022", "2023", "请教", "介绍",
        "怎样", "吗", "么", "？", "?", "呢",
        "评价", "搜索", "百度", "谷歌", "bing", "谁是", "上网"
    ]

    def pre_check(self):
        try:
            from duckduckgo_search import DDGS
            from llmkira.sdk.filter import Sublimate
            return True
        except ImportError as e:
            logger.warning(f"plugin:package <duckduckgo_search> not found,please install it first:{e}")
            return False

    def func_message(self, message_text):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(self, platform, task, receiver, reason):
        try:
            await Task(queue=platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=TaskHeader.Meta(
                        callback_forward=True,
                        callback=TaskHeader.Meta.Callback(
                            role="function",
                            name=__plugin_name__
                        ),
                    ),
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text=f"🍖 {__plugin_name__}操作失败了！原因：{reason}"
                        )
                    ]
                )
            )
        except Exception as e:
            logger.error(e)

    @staticmethod
    async def llm_task(task, task_desc, raw_data):
        _submanager = SubManager(user_id=f"{task.sender.platform}:{task.sender.user_id}")
        driver = _submanager.llm_driver  # 由发送人承担接受者的成本
        model_name = os.getenv("OPENAI_API_MODEL", "gpt-3.5-turbo-0613")
        endpoint = openai.Openai(
            config=driver,
            model=model_name,
            messages=Message.create_task_message_list(
                task_desc=task_desc,
                refer=raw_data
            ),
        )
        # 调用Openai
        result = await endpoint.create()
        await _submanager.add_cost(
            cost=UserInfo.Cost(token_usage=result.usage.total_tokens, token_uuid=driver.uuid, model_name=model_name)
        )
        return result.default_message.content

    async def callback(self, sign: str, task: TaskHeader):
        if sign == "reply":
            chain: Chain = await CHAIN_MANAGER.get_task(user_id=str(task.receiver.user_id))
            if chain:
                logger.info(f"{__plugin_name__}:chain callback locate in {sign} be sent")
                await Task(queue=chain.address).send_task(task=chain.arg)
            return True
        else:
            return False

    async def run(self, task: TaskHeader, receiver: TaskHeader.Location, arg, **kwargs):
        """
        处理message，返回message
        """
        try:
            _set = Search.parse_obj(arg)
            _search_result = search_on_duckduckgo(_set.keywords)

            # META
            _meta = task.task_meta.child(__plugin_name__)
            _meta.callback_forward = True
            _meta.callback_forward_reprocess = True
            _meta.callback = TaskHeader.Meta.Callback(
                role="function",
                name=__plugin_name__
            )
            await Task(queue=receiver.platform).send_task(
                task=TaskHeader(
                    sender=task.sender,  # 继承发送者
                    receiver=receiver,  # 因为可能有转发，所以可以单配
                    task_meta=_meta,
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text=_search_result
                        )
                    ]
                )
            )
        except Exception as e:
            await self.failed(platform=receiver.platform, task=task, receiver=receiver, reason="搜索失败了！")
            logger.exception(e)


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Search/validate uncertain/unknownEvents/Meme fact on google.com",
    usage="search <keywords>",
    openapi_version="20231013",
    function={
        FuncPair(function=search, tool=SearchTool)
    },
)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午9:54
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import asyncio

import aio_pika
from aio_pika import Message, DeliveryMode
from aiormq import DeliveryError
from loguru import logger
from pamqp.commands import Basic

from .schema import TaskHeader, RabbitMQSetting

EXPIRATION_SECOND = 60 * 5  # 5min
QUEUE_MAX_LENGTH = 120
X_OVERFLOW = "reject-publish"  # 拒绝
CONSUMER_PREFETCH_COUNT = 12  # 消息流控
QUEUE_ARGUMENTS = {
    "x-max-length": QUEUE_MAX_LENGTH,  # 上限
    # "message-ttl": EXPIRATION_SECOND * 1000,
    "x-overflow": X_OVERFLOW  # 拒绝
}


class Task(object):
    def __init__(self, queue: str):
        """
        :param queue: 队列名字
        """
        self.queue_name = queue
        self.amqp_url = RabbitMQSetting.amqp_dsn

    async def send_task(self, task: TaskHeader):
        connection = await aio_pika.connect_robust(self.amqp_url)
        async with connection:
            channel = await connection.channel(
                publisher_confirms=True  # 消息确认
            )
            # 通道超时 2s
            # await channel.initialize(timeout=2000)
            # Creating a message
            message = Message(
                task.json().encode("utf-8"),
                delivery_mode=DeliveryMode.PERSISTENT,
                expiration=EXPIRATION_SECOND
            )
            # Declaring queue
            try:
                await channel.declare_queue(
                    self.queue_name,
                    arguments=QUEUE_ARGUMENTS,
                    durable=True  # 持续化
                )
            except Exception as e:
                logger.error(
                    f"[5941163]Rabbitmq Queue param validation error, try deleting the abnormal queue manually and retrying. "
                    f"\n--error {e}"
                    f"\n--help |web: <database_ip>:15672/#/ |shell: `rabbitmqctl delete_queue {self.queue_name}`"
                )
                raise e
            # Sending the message
            try:
                confirmation = await channel.default_exchange.publish(
                    message,
                    routing_key=self.queue_name,
                    timeout=10
                )
            except DeliveryError as e:
                logger.error(f"[502231]Task Delivery failed with exception: {e.reason}")
                return False, "🔭 Sorry,failure to reach the control centre, possibly i reach design limit 🧯"
            except TimeoutError:
                logger.error(f"[528532]Task Delivery timeout")
                return False, "🔭 Sorry, failure to reach the control centre, ⏱ timeout"
            else:
                if not isinstance(confirmation, Basic.Ack):
                    logger.error(f"[552123]Task Delivery failed with confirmation")
                    return False, "🔭 Sorry, failure to reach the control centre, sender error"
                else:
                    logger.info("[621132]Task sent success")
                    return True, "Task sent success"

    async def consuming_task(self, func: callable):
        connection = await aio_pika.connect_robust(self.amqp_url)
        async with connection:
            # Creating a channel
            channel = await connection.channel()
            # 通道超时 1s
            # await channel.initialize(timeout=1000)
            await channel.set_qos(prefetch_count=CONSUMER_PREFETCH_COUNT)  # 消息流控

            # Declaring queue
            try:
                queue = await channel.declare_queue(
                    self.queue_name,
                    arguments=QUEUE_ARGUMENTS,
                    durable=True  # 持续化
                )
            except Exception as e:
                logger.error(
                    f"[502231]Rabbitmq Queue parameter validation failed, try deleting the abnormal queue manually and retrying. "
                    f"\n--error {e}"
                    f"\n--help |web: <database_ip>:15672/#/ |shell: `rabbitmqctl delete_queue {self.queue_name}`"
                )
                raise e
            await queue.consume(func)
            await asyncio.Future()  # run forever



# -*- coding: utf-8 -*-
# @Time    : 2023/9/4 下午10:49
# @Author  : sudoskys
# @File    : reduce.py
# @Software: PyCharm
from typing import List, Any

import cjieba
import numpy as np
from sklearn.cluster import Birch
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfTransformer


class Cluster(object):

    def __init__(self):
        self.weight = None
        self.cluster = None
        self.title_dict: dict = {}

    def init(self, sentence_list):
        corpus = []
        self.title_dict = {}
        index = 0
        for line in sentence_list:
            title = line.strip()
            self.title_dict[index] = title
            output = ' '.join(['%s' % x for x in list(cjieba.cut(title, cut_all=False))]).encode('utf-8')  # 空格拼接
            index += 1
            corpus.append(output.strip())
        _vectorizer = CountVectorizer()
        transformer = TfidfTransformer()
        tfidf = transformer.fit_transform(_vectorizer.fit_transform(corpus))
        word = _vectorizer.get_feature_names_out()
        self.weight = tfidf.toarray()

    def birch_cluster(self, sentence_list: List[str], threshold: int = 0.6) -> None:
        self.init(sentence_list=sentence_list)
        self.cluster = Birch(threshold=threshold, n_clusters=None)
        self.cluster.fit_predict(self.weight)

    def build(self) -> List[Any]:
        # self.cluster.labels_  对应 类别 {index: 类别} 类别值int值 相同值代表同一类
        # cluster_dict key为Birch聚类后的每个类，value为 title对应的index
        cluster_dict = {}
        for index, value in enumerate(self.cluster.labels_):
            if value not in cluster_dict:
                cluster_dict[value] = [index]
            else:
                cluster_dict[value].append(index)
        # print("-----before cluster Birch count title:", len(self.title_dict))
        # result_dict key为Birch聚类后距离中心点最近的title，value为sum_similar求和
        result_dict = {}
        for _index in cluster_dict.values():
            latest_index = _index[0]
            similar_num = len(_index)
            if len(_index) >= 2:
                min_s = np.sqrt(np.sum(np.square(
                    self.weight[_index[0]] - self.cluster.subcluster_centers_[self.cluster.labels_[_index[0]]])))
                for index in _index:
                    s = np.sqrt(np.sum(
                        np.square(self.weight[index] - self.cluster.subcluster_centers_[self.cluster.labels_[index]])))
                    if s < min_s:
                        min_s = s
                        latest_index = index
            title = self.title_dict[latest_index]
            result_dict[title] = similar_num
        return list(result_dict.keys())
